#include <iostream> 
#include <fstream>
#include <utility>
#include <string>
#include <sstream>
#include <stdlib.h>
#include <math.h>
#ifndef RASTER_H
#define RASTER_H
// Raster Class
template <typename T>
class Raster
{
	public:
		  // constructor
		Raster(unsigned wd,unsigned ht, double xmin, double xmax, double ymin, double ymax) : nWd(wd), nHt(ht), pAr(0)
		{
			if(wd > 0 && ht > 0)
				pAr = new T[wd * ht];
			LON_MIN = xmin;
			LON_MAX = xmax;
			LAT_MIN = ymin;
			LAT_MAX = ymax;
			LON_RANGE = xmax-xmin;
			LAT_RANGE = ymax-ymin;
		}

		// destructor
		~Raster()
		{
			delete[] pAr;
		}

		// indexing (parenthesis operator)
		//  two of them (for const correctness)

		const T& operator () (unsigned x,unsigned y) const
		{  return pAr[ y*nWd + x ];   }

		T& operator () (unsigned x,unsigned y)
		{  return pAr[ y*nWd + x ];   }

		// get dims
		unsigned GetWd() const { return nWd; }
		unsigned GetHt() const { return nHt; }

		void init(T def)
		{
			for( int i=0; i<nWd*nHt; i++)
				pAr[i] = def;
		}

		void loadData(const string& fname)
		{
			cout << "Loading raster data..." << endl;
			ifstream fid (fname.c_str(),ifstream::in);
			string tline;
			
			for (int y=0; y < nHt ;y++){
				getline(fid,tline);
				stringstream parse_line(tline);
				for (int x=0; x < nWd; x++){
					parse_line>>pAr[y*nWd + x];
				}
			}
		}

		int getX(double lon)
		{
			int x = (int)((lon-LON_MIN)/LON_RANGE*nWd);
			return x;
		}

		int getY(double lat)
		{
			int y = nHt - (int)((lat-LAT_MIN)/LAT_RANGE*nHt);
			return y;
		}

		int inPoly(double lon, double lat)
		{
			int x;
			int y;
			x=(int)((lon-LON_MIN)/LON_RANGE*nWd);
			y= nHt - (int)((lat-LAT_MIN)/LAT_RANGE*nHt);
			if ( x<nWd && y<nHt && x>-1 && y>-1)
			{
				return pAr[ y*nWd + x ];
			}
			else{
				return 0;
			}
		}
		// private data members
	private:
		unsigned nWd;
		unsigned nHt;
		double LON_MIN;
		double LON_MAX;
		double LAT_MIN;
		double LAT_MAX;
		double LON_RANGE;
		double LAT_RANGE;
		T*       pAr;

		// to prevent unwanted copying:
		Raster(const Raster<T>&);
		Raster& operator = (const Raster<T>&);
};
#endif
