#include <iostream> 
#include <fstream>
#include <utility>
#include <string>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/random.hpp>
#include <boost/random.hpp>
#include <boost/graph/astar_search.hpp>
#include <vector>
#include <ctime>
#include <thread>
#include <map>
#include <unordered_map>
#include <mutex>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sstream>
#include <libpq-fe.h>
#include "RoadNetwork.hpp"
#include "OD.hpp"
#include "Routing.hpp"
#include "Raster.hpp"

using namespace std;

int NUM_THREADS;
int NUM_INCREMENTS;
string NETWORK_TABLE;
bool DEBUG = false;
bool UPDATE = false;
bool SINGLE = false;
bool REMOTE = false;
string FILE_NODES;
string FILE_EDGES;
string FILE_OUT;
string FILE_OD;
string FILE_RASTER;
mutex vol_mutex;

// MAIN PROGRAM
int main(int argc, char* argv[])
{ 
    
    int c, s, t;
    extern char *optarg;
    while ((c = getopt(argc, argv, "s:t:i:b:du:r:n:e:f:o:p:")) != -1)
        switch (c) {
            case 's':
                SINGLE=true;
                s = atoi(optarg);
                break;
            case 't':
                t = atoi(optarg);
                break;
            case 'i':
                NUM_INCREMENTS = atoi(optarg);
                break;
            case 'b':
                NUM_THREADS = atoi(optarg);
                break;
            case 'd':
                DEBUG = true;
                break;
            case 'u':
                UPDATE = true;
                NETWORK_TABLE = optarg;
                break;
            case 'r':
                REMOTE = true;
                NETWORK_TABLE = optarg;
                break;
            case 'n':
                FILE_NODES = optarg;
                break;
            case 'e':
                FILE_EDGES = optarg;
                break;
            case 'f':
                FILE_OD = optarg;
                break;
            case 'o':
                FILE_OUT = optarg;
                break;
            case 'p':
                FILE_RASTER = optarg;
                break;
            }

    // read in network
    RoadNetwork g;
    if(!REMOTE){ g = createNetworkFromFile(FILE_NODES, FILE_EDGES);}
    else { g = createNetworkFromDatabase(NETWORK_TABLE);}

    // create a location map for nodes
    NodeMap locations = boost::get(boost::vertex_bundle, g);

    // load census raster
    Raster<int> raster_map(10734, 5000,-73.5081, -69.8586, 41.1871, 42.8871);
    raster_map.init(0);
    raster_map.loadData(FILE_RASTER);

    if (SINGLE)
    {
        singleRoute(g, &locations,s,t);
        exit(0);
    }
    // read in OD Pairs
    OD od (NUM_INCREMENTS,NUM_THREADS);
    od.fromFile(FILE_OD);
    cout << "Number of OD Pairs: " << od.pairs.size() << endl;
    cout << "Number of OD Pairs per Batch: " << od.pairs.size()/NUM_THREADS << endl;
    cout << "Number of Trips: "<< od.totalTrips() << endl;
    /*
    // dummy OD for testing
    vector<ODPair> odp;
    odp.push_back( ODPair(1,2,3) );
    odp.push_back( ODPair(4,5,6) );
    odp.push_back( ODPair(7,8,9) );
    OD od = OD(3,3);
    od.pairs = odp;
    */
    clock_t globalStartTime = clock();

    // make a map to store road volume
    EdgeMap volume;
    boost::graph_traits<RoadNetwork>::edge_iterator ei, ei_end;
    // make map to store census tracts associated with roads
    EdgeTractMap tractMap;

    int total_trips = 0;
    for( int i=0; i < NUM_INCREMENTS; i++)
    {
        cout << "Starting increment " << i << endl;
        // init map with zero volume on every edge for this increment
        total_trips = 0;
        for (boost::tie(ei, ei_end) = boost::edges(g); ei != ei_end; ++ei)
            volume[*ei] = 0;
        // generate threads and do routing
        vector<thread> threads;
        for (int t=0; t < NUM_THREADS ; t++)
            threads.push_back( thread(&batchRoute, &g, &volume, &tractMap, &od, &locations, t, &raster_map) );
        // wait for all the batches to finish running
        for (auto& th : threads) th.join();
        // update the travel times on the roads
        updateTravelTimes(&g, &volume);
        // count the total number of segments traveled on
        for (boost::tie(ei, ei_end) = boost::edges(g); ei != ei_end; ++ei)
            total_trips += volume[*ei];
        cout << "Total Volume in increment: " << total_trips << endl;
    } 
    cout << "Total Time: " << double( clock() - globalStartTime ) / (double)CLOCKS_PER_SEC / (double)NUM_THREADS << " seconds." << endl;
    cout << "Done" << endl;
    printEdges(FILE_OUT, &g, &tractMap);

    if (UPDATE)
    {
        cout << "Updating database...";
        updateDatabaseNetwork(NETWORK_TABLE, &g);
        string usage_table = NETWORK_TABLE.append("_usage");
        updateDatabaseUsage(usage_table, &g, &tractMap);
        cout << "DONE!" << endl;
    }
    return 0;
}
